@@ -1,16 +1,23 @@
 import logging
 import os
 import signal
+import sys
 import time
 
-from patroni.daemon import AbstractPatroniDaemon, abstract_main
+from argparse import Namespace
+from typing import Any, Dict, Optional, TYPE_CHECKING
+
+from patroni.daemon import AbstractPatroniDaemon, abstract_main, get_base_arg_parser
+
+if TYPE_CHECKING:  # pragma: no cover
+    from .config import Config
 
 logger = logging.getLogger(__name__)
 
 
 class Patroni(AbstractPatroniDaemon):
 
-    def __init__(self, config):
+    def __init__(self, config: 'Config') -> None:
         from patroni.api import RestApiServer
         from patroni.dcs import get_dcs
         from patroni.ha import Ha
@@ -23,19 +30,22 @@ class Patroni(AbstractPatroniDaemon):
 
         self.version = __version__
         self.dcs = get_dcs(self.config)
+        self.request = PatroniRequest(self.config, True)
+
+        self.ensure_unique_name()
+
         self.watchdog = Watchdog(self.config)
         self.load_dynamic_configuration()
 
         self.postgresql = Postgresql(self.config['postgresql'])
         self.api = RestApiServer(self, self.config['restapi'])
-        self.request = PatroniRequest(self.config, True)
         self.ha = Ha(self)
 
         self.tags = self.get_tags()
         self.next_run = time.time()
-        self.scheduled_restart = {}
+        self.scheduled_restart: Dict[str, Any] = {}
 
-    def load_dynamic_configuration(self):
+    def load_dynamic_configuration(self) -> None:
         from patroni.exceptions import DCSError
         while True:
             try:
@@ -53,19 +63,36 @@ class Patroni(AbstractPatroniDaemon):
                 logger.warning('Can not get cluster from dcs')
                 time.sleep(5)
 
-    def get_tags(self):
+    def ensure_unique_name(self) -> None:
+        """A helper method to prevent splitbrain from operator naming error."""
+        from patroni.dcs import Member
+
+        cluster = self.dcs.get_cluster()
+        if not cluster:
+            return
+        member = cluster.get_member(self.config['name'], False)
+        if not isinstance(member, Member):
+            return
+        try:
+            _ = self.request(member, endpoint="/liveness")
+            logger.fatal("Can't start; there is already a node named '%s' running", self.config['name'])
+            sys.exit(1)
+        except Exception:
+            return
+
+    def get_tags(self) -> Dict[str, Any]:
         return {tag: value for tag, value in self.config.get('tags', {}).items()
                 if tag not in ('clonefrom', 'nofailover', 'noloadbalance', 'nosync') or value}
 
     @property
-    def nofailover(self):
+    def nofailover(self) -> bool:
         return bool(self.tags.get('nofailover', False))
 
     @property
-    def nosync(self):
+    def nosync(self) -> bool:
         return bool(self.tags.get('nosync', False))
 
-    def reload_config(self, sighup=False, local=False):
+    def reload_config(self, sighup: bool = False, local: Optional[bool] = False) -> None:
         try:
             super(Patroni, self).reload_config(sighup, local)
             if local:
@@ -87,7 +114,7 @@ class Patroni(AbstractPatroniDaemon):
     def noloadbalance(self):
         return bool(self.tags.get('noloadbalance', False))
 
-    def schedule_next_run(self):
+    def schedule_next_run(self) -> None:
         self.next_run += self.dcs.loop_wait
         current_time = time.time()
         nap_time = self.next_run - current_time
@@ -100,12 +127,12 @@ class Patroni(AbstractPatroniDaemon):
         elif self.ha.watch(nap_time):
             self.next_run = time.time()
 
-    def run(self):
+    def run(self) -> None:
         self.api.start()
         self.next_run = time.time()
         super(Patroni, self).run()
 
-    def _run_cycle(self):
+    def _run_cycle(self) -> None:
         logger.info(self.ha.run_cycle())
 
         if self.dcs.cluster and self.dcs.cluster.config and self.dcs.cluster.config.data \
@@ -117,7 +144,7 @@ class Patroni(AbstractPatroniDaemon):
 
         self.schedule_next_run()
 
-    def _shutdown(self):
+    def _shutdown(self) -> None:
         try:
             self.api.shutdown()
         except Exception:
@@ -128,26 +155,47 @@ class Patroni(AbstractPatroniDaemon):
             logger.exception('Exception during Ha.shutdown')
 
 
-def patroni_main():
+def patroni_main(configfile: str) -> None:
     from multiprocessing import freeze_support
-    from patroni.validator import schema
 
     freeze_support()
-    abstract_main(Patroni, schema)
+    abstract_main(Patroni, configfile)
 
 
-def main():
+def process_arguments() -> Namespace:
+    parser = get_base_arg_parser()
+    parser.add_argument('--validate-config', action='store_true', help='Run config validator and exit')
+    args = parser.parse_args()
+
+    if args.validate_config:
+        from patroni.validator import schema
+        from patroni.config import Config, ConfigParseError
+
+        try:
+            Config(args.configfile, validator=schema)
+            sys.exit()
+        except ConfigParseError as e:
+            sys.exit(e.value)
+
+    return args
+
+
+def main() -> None:
+    from patroni import check_psycopg
+
+    args = process_arguments()
+
+    check_psycopg()
+
     if os.getpid() != 1:
-        from patroni import check_psycopg
-
-        check_psycopg()
-        return patroni_main()
+        return patroni_main(args.configfile)
 
     # Patroni started with PID=1, it looks like we are in the container
+    from types import FrameType
     pid = 0
 
     # Looks like we are in a docker, so we will act like init
-    def sigchld_handler(signo, stack_frame):
+    def sigchld_handler(signo: int, stack_frame: Optional[FrameType]) -> None:
         try:
             while True:
                 ret = os.waitpid(-1, os.WNOHANG)
@@ -158,7 +206,7 @@ def main():
         except OSError:
             pass
 
-    def passtochild(signo, stack_frame):
+    def passtochild(signo: int, stack_frame: Optional[FrameType]):
         if pid:
             os.kill(pid, signo)
 
@@ -173,7 +221,7 @@ def main():
     signal.signal(signal.SIGTERM, passtochild)
 
     import multiprocessing
-    patroni = multiprocessing.Process(target=patroni_main)
+    patroni = multiprocessing.Process(target=patroni_main, args=(args.configfile,))
     patroni.start()
     pid = patroni.pid
     patroni.join()
