@@ -11,6 +11,7 @@ from mock import Mock, PropertyMock, patch
 from socketserver import ThreadingMixIn
 
 from patroni.api import RestApiHandler, RestApiServer
+from patroni.config import GlobalConfig
 from patroni.dcs import ClusterConfig, Member
 from patroni.ha import _MemberStatus
 from patroni.utils import tzutc
@@ -28,7 +29,8 @@ class MockPostgresql(object):
     name = 'test'
     state = 'running'
     role = 'primary'
-    server_version = '999999'
+    server_version = 90625
+    major_version = 90600
     sysid = 'dummysysid'
     scope = 'dummy'
     pending_restart = True
@@ -54,6 +56,10 @@ class MockPostgresql(object):
     def is_running():
         return True
 
+    @staticmethod
+    def replication_state_from_parameters(*args):
+        return 'streaming'
+
 
 class MockWatchdog(object):
     is_healthy = False
@@ -116,10 +122,6 @@ class MockHa(object):
     def is_paused():
         return True
 
-    @staticmethod
-    def is_standby_cluster():
-        return False
-
 
 class MockLogger(object):
 
@@ -128,10 +130,16 @@ class MockLogger(object):
     records_lost = 1
 
 
+class MockConfig(object):
+
+    def get_global_config(self, _):
+        return GlobalConfig({})
+
+
 class MockPatroni(object):
 
     ha = MockHa()
-    config = Mock()
+    config = MockConfig()
     postgresql = ha.state_handler
     dcs = Mock()
     logger = MockLogger()
@@ -185,7 +193,8 @@ class TestRestApiHandler(unittest.TestCase):
     def test_do_GET(self):
         MockPatroni.dcs.cluster.last_lsn = 20
         MockPatroni.dcs.cluster.sync.members = [MockPostgresql.name]
-        MockRestApiServer(RestApiHandler, 'GET /replica')
+        with patch.object(GlobalConfig, 'is_synchronous_mode', PropertyMock(return_value=True)):
+            MockRestApiServer(RestApiHandler, 'GET /replica')
         MockRestApiServer(RestApiHandler, 'GET /replica?lag=1M')
         MockRestApiServer(RestApiHandler, 'GET /replica?lag=10MB')
         MockRestApiServer(RestApiHandler, 'GET /replica?lag=10485760')
@@ -207,7 +216,7 @@ class TestRestApiHandler(unittest.TestCase):
         with patch.object(MockHa, 'is_leader', Mock(return_value=True)):
             MockRestApiServer(RestApiHandler, 'GET /replica')
             MockRestApiServer(RestApiHandler, 'GET /read-only-sync')
-            with patch.object(MockHa, 'is_standby_cluster', Mock(return_value=True)):
+            with patch.object(GlobalConfig, 'is_standby_cluster', Mock(return_value=True)):
                 MockRestApiServer(RestApiHandler, 'GET /standby_leader')
         MockPatroni.dcs.cluster = None
         with patch.object(RestApiHandler, 'get_postgresql_status', Mock(return_value={'role': 'primary'})):
@@ -215,9 +224,10 @@ class TestRestApiHandler(unittest.TestCase):
         with patch.object(MockHa, 'restart_scheduled', Mock(return_value=True)):
             MockRestApiServer(RestApiHandler, 'GET /primary')
         self.assertIsNotNone(MockRestApiServer(RestApiHandler, 'GET /primary'))
-        with patch.object(RestApiServer, 'query', Mock(return_value=[('', 1, '', '', '', '', False, '')])):
+        with patch.object(RestApiServer, 'query', Mock(return_value=[('', 1, '', '', '', '', False, None, None, '')])):
             self.assertIsNotNone(MockRestApiServer(RestApiHandler, 'GET /patroni'))
-        with patch.object(MockHa, 'is_standby_cluster', Mock(return_value=True)):
+        with patch.object(GlobalConfig, 'is_standby_cluster', Mock(return_value=True)),\
+                patch.object(GlobalConfig, 'is_paused', Mock(return_value=True)):
             MockRestApiServer(RestApiHandler, 'GET /standby_leader')
 
         # test tags
@@ -393,8 +403,8 @@ class TestRestApiHandler(unittest.TestCase):
         with patch.object(MockHa, 'is_failsafe_mode', Mock(return_value=False), create=True):
             self.assertIsNotNone(MockRestApiServer(RestApiHandler, 'POST /failsafe HTTP/1.0' + self._authorization))
         with patch.object(MockHa, 'is_failsafe_mode', Mock(return_value=True), create=True):
-            self.assertIsNotNone(MockRestApiServer(RestApiHandler, 'POST /failsafe HTTP/1.0' + self._authorization +
-                                                                   '\nContent-Length: 9\n\n{"a":"b"}'))
+            self.assertIsNotNone(MockRestApiServer(RestApiHandler, 'POST /failsafe HTTP/1.0' + self._authorization
+                                                   + '\nContent-Length: 9\n\n{"a":"b"}'))
 
     @patch.object(MockPatroni, 'sighup_handler', Mock())
     def test_do_POST_reload(self):
@@ -405,9 +415,7 @@ class TestRestApiHandler(unittest.TestCase):
     def test_do_POST_sigterm(self):
         self.assertIsNotNone(MockRestApiServer(RestApiHandler, 'POST /sigterm HTTP/1.0' + self._authorization))
 
-    @patch.object(MockPatroni, 'dcs')
-    def test_do_POST_restart(self, mock_dcs):
-        mock_dcs.get_cluster.return_value.is_paused.return_value = False
+    def test_do_POST_restart(self):
         request = 'POST /restart HTTP/1.0' + self._authorization
         self.assertIsNotNone(MockRestApiServer(RestApiHandler, request))
 
@@ -449,12 +457,12 @@ class TestRestApiHandler(unittest.TestCase):
                 request = make_request(role='primary', postgres_version='9.5.2')
                 MockRestApiServer(RestApiHandler, request)
 
-        mock_dcs.get_cluster.return_value.is_paused.return_value = True
-        MockRestApiServer(RestApiHandler, make_request(schedule='2016-08-42 12:45TZ+1', role='primary'))
-        # Valid timeout
-        MockRestApiServer(RestApiHandler, make_request(timeout='60s'))
-        # Invalid timeout
-        MockRestApiServer(RestApiHandler, make_request(timeout='42towels'))
+        with patch.object(GlobalConfig, 'is_paused', PropertyMock(return_value=True)):
+            MockRestApiServer(RestApiHandler, make_request(schedule='2016-08-42 12:45TZ+1', role='primary'))
+            # Valid timeout
+            MockRestApiServer(RestApiHandler, make_request(timeout='60s'))
+            # Invalid timeout
+            MockRestApiServer(RestApiHandler, make_request(timeout='42towels'))
 
     def test_do_DELETE_restart(self):
         for retval in (True, False):
@@ -471,10 +479,7 @@ class TestRestApiHandler(unittest.TestCase):
         mock_dcs.get_cluster.return_value.failover = None
         self.assertIsNotNone(MockRestApiServer(RestApiHandler, request))
 
-    @patch.object(MockPatroni, 'dcs')
-    def test_do_POST_reinitialize(self, mock_dcs):
-        cluster = mock_dcs.get_cluster.return_value
-        cluster.is_paused.return_value = False
+    def test_do_POST_reinitialize(self):
         request = 'POST /reinitialize HTTP/1.0' + self._authorization + '\nContent-Length: 15\n\n{"force": true}'
         MockRestApiServer(RestApiHandler, request)
         with patch.object(MockHa, 'reinitialize', Mock(return_value=None)):
@@ -492,8 +497,6 @@ class TestRestApiHandler(unittest.TestCase):
     def test_do_POST_switchover(self, dcs):
         dcs.loop_wait = 10
         cluster = dcs.get_cluster.return_value
-        cluster.is_synchronous_mode.return_value = False
-        cluster.is_paused.return_value = False
 
         post = 'POST /switchover HTTP/1.0' + self._authorization + '\nContent-Length: '
 
@@ -507,20 +510,22 @@ class TestRestApiHandler(unittest.TestCase):
 
         request = post + '25\n\n{"leader": "postgresql1"}'
 
-        cluster.is_paused.return_value = True
-        MockRestApiServer(RestApiHandler, request)
-
-        cluster.is_paused.return_value = False
-        for cluster.is_synchronous_mode.return_value in (True, False):
+        with patch.object(GlobalConfig, 'is_paused', PropertyMock(return_value=True)):
             MockRestApiServer(RestApiHandler, request)
 
+        for is_synchronous_mode in (True, False):
+            with patch.object(GlobalConfig, 'is_synchronous_mode', PropertyMock(return_value=is_synchronous_mode)):
+                MockRestApiServer(RestApiHandler, request)
+
         cluster.leader.name = 'postgresql2'
         request = post + '53\n\n{"leader": "postgresql1", "candidate": "postgresql2"}'
         MockRestApiServer(RestApiHandler, request)
 
         cluster.leader.name = 'postgresql1'
-        for cluster.is_synchronous_mode.return_value in (True, False):
-            MockRestApiServer(RestApiHandler, request)
+        cluster.sync.matches.return_value = False
+        for is_synchronous_mode in (True, False):
+            with patch.object(GlobalConfig, 'is_synchronous_mode', PropertyMock(return_value=is_synchronous_mode)):
+                MockRestApiServer(RestApiHandler, request)
 
         cluster.members = [Member(0, 'postgresql0', 30, {'api_url': 'http'}),
                            Member(0, 'postgresql2', 30, {'api_url': 'http'})]
@@ -554,7 +559,8 @@ class TestRestApiHandler(unittest.TestCase):
         request = post + '103\n\n{"leader": "postgresql1", "member": "postgresql2",' +\
                          ' "scheduled_at": "6016-02-15T18:13:30.568224+01:00"}'
         MockRestApiServer(RestApiHandler, request)
-        with patch.object(MockPatroni, 'dcs') as d:
+        with patch.object(GlobalConfig, 'is_paused', PropertyMock(return_value=True)),\
+                patch.object(MockPatroni, 'dcs') as d:
             d.manual_failover.return_value = False
             MockRestApiServer(RestApiHandler, request)
 
@@ -570,13 +576,11 @@ class TestRestApiHandler(unittest.TestCase):
         # Invalid date
         self.assertIsNotNone(MockRestApiServer(RestApiHandler, request + '2010-02-29T18:13:30.568224+01:00"}'))
 
-    @patch.object(MockPatroni, 'dcs', Mock())
     def test_do_POST_failover(self):
         post = 'POST /failover HTTP/1.0' + self._authorization + '\nContent-Length: '
         MockRestApiServer(RestApiHandler, post + '14\n\n{"leader":"1"}')
         MockRestApiServer(RestApiHandler, post + '37\n\n{"candidate":"2","scheduled_at": "1"}')
 
-    @patch.object(MockPatroni, 'dcs', Mock())
     @patch.object(MockHa, 'is_leader', Mock(return_value=True))
     def test_do_POST_citus(self):
         post = 'POST /citus HTTP/1.0' + self._authorization + '\nContent-Length: '
@@ -622,24 +626,38 @@ class TestRestApiServer(unittest.TestCase):
         try:
             raise Exception()
         except Exception:
-            self.assertIsNone(MockRestApiServer.handle_error(None, ('127.0.0.1', 55555)))
+            self.assertIsNone(self.srv.handle_error(None, ('127.0.0.1', 55555)))
 
     @patch.object(HTTPServer, '__init__', Mock(side_effect=socket.error))
     def test_socket_error(self):
         self.assertRaises(socket.error, MockRestApiServer, Mock(), '', {'listen': '*:8008'})
 
+    def __create_socket(self):
+        sock = socket.socket()
+        try:
+            import ssl
+            ctx = ssl.create_default_context(ssl.Purpose.SERVER_AUTH)
+            ctx.check_hostname = False
+            sock = ctx.wrap_socket(sock=sock)
+            sock.do_handshake = Mock()
+            sock.unwrap = Mock(side_effect=Exception)
+        except Exception:
+            pass
+        return sock
+
     @patch.object(ThreadingMixIn, 'process_request_thread', Mock())
     def test_process_request_thread(self):
-        self.srv.process_request_thread(Mock(), '2')
+        self.srv.process_request_thread(self.__create_socket(), ('2', 54321))
 
     @patch.object(MockRestApiServer, 'process_request', Mock(side_effect=RuntimeError))
     @patch.object(MockRestApiServer, 'get_request')
     def test_process_request_error(self, mock_get_request):
-        mock_request = Mock()
-        mock_request.unwrap.side_effect = Exception
-        mock_get_request.return_value = (mock_request, ('127.0.0.1', 55555))
+        mock_get_request.return_value = (self.__create_socket(), ('127.0.0.1', 55555))
         self.srv._handle_request_noblock()
 
     @patch('ssl._ssl._test_decode_cert', Mock())
     def test_reload_local_certificate(self):
         self.assertTrue(self.srv.reload_local_certificate())
+
+    def test_get_certificate_serial_number(self):
+        self.assertIsNone(self.srv.get_certificate_serial_number())
