@@ -5,14 +5,13 @@ import io
 
 from mock import MagicMock, Mock, patch
 from patroni.config import Config, ConfigParseError
-from six.moves import builtins
 
 
 class TestConfig(unittest.TestCase):
 
     @patch('os.path.isfile', Mock(return_value=True))
     @patch('json.load', Mock(side_effect=Exception))
-    @patch.object(builtins, 'open', MagicMock())
+    @patch('builtins.open', MagicMock())
     def setUp(self):
         sys.argv = ['patroni.py']
         os.environ[Config.PATRONI_CONFIG_VARIABLE] = 'restapi: {}\npostgresql: {data_dir: foo}'
@@ -20,10 +19,9 @@ class TestConfig(unittest.TestCase):
 
     def test_set_dynamic_configuration(self):
         with patch.object(Config, '_build_effective_configuration', Mock(side_effect=Exception)):
-            self.assertIsNone(self.config.set_dynamic_configuration({'foo': 'bar'}))
-        self.assertTrue(self.config.set_dynamic_configuration({'synchronous_mode': True,
-                                                               'standby_cluster': {}, 'master_start_timeout': 1}))
-        self.assertEqual(self.config.get('primary_start_timeout'), 1)
+            self.assertFalse(self.config.set_dynamic_configuration({'foo': 'bar'}))
+        self.assertTrue(self.config.set_dynamic_configuration({'standby_cluster': {}, 'postgresql': {
+            'parameters': {'cluster_name': 1, 'wal_keep_size': 1, 'track_commit_timestamp': 1, 'wal_level': 1}}}))
 
     def test_reload_local_configuration(self):
         os.environ.update({
@@ -61,6 +59,7 @@ class TestConfig(unittest.TestCase):
             'PATRONI_KUBERNETES_LABELS': 'a: b: c',
             'PATRONI_KUBERNETES_SCOPE_LABEL': 'a',
             'PATRONI_KUBERNETES_PORTS': '[{"name": "postgresql"}]',
+            'PATRONI_KUBERNETES_RETRIABLE_HTTP_CODES': '401',
             'PATRONI_ZOOKEEPER_HOSTS': "'host1:2181','host2:2181'",
             'PATRONI_EXHIBITOR_HOSTS': 'host1,host2',
             'PATRONI_EXHIBITOR_PORT': '8181',
@@ -71,7 +70,8 @@ class TestConfig(unittest.TestCase):
             'PATRONI_REPLICATION_USERNAME': 'replicator',
             'PATRONI_REPLICATION_PASSWORD': 'rep-pass',
             'PATRONI_admin_PASSWORD': 'admin',
-            'PATRONI_admin_OPTIONS': 'createrole,createdb'
+            'PATRONI_admin_OPTIONS': 'createrole,createdb',
+            'PATRONI_POSTGRESQL_BIN_POSTGRES': 'sergtsop'
         })
         config = Config('postgres0.yml')
         with patch.object(Config, '_load_config_file', Mock(return_value={'restapi': {}})):
@@ -136,7 +136,7 @@ class TestConfig(unittest.TestCase):
                     new-attr: True
                     ''')
 
-        with patch.object(builtins, 'open', MagicMock(side_effect=open_mock)):
+        with patch('builtins.open', MagicMock(side_effect=open_mock)):
             config = Config('postgres0')
             self.assertEqual(config._local_configuration,
                              {'test': False, 'test2': {'child-1': 'somestring', 'child-2': 10},
