@@ -20,8 +20,8 @@ class MockThread(object):
 def mock_cancellable_call(*args, **kwargs):
     communicate = kwargs.pop('communicate', None)
     if isinstance(communicate, dict):
-        communicate.update(stdout=b'', stderr=b'pg_rewind: error: could not open file ' +
-                                              b'"data/postgresql0/pg_xlog/000000010000000000000003": No such file')
+        communicate.update(stdout=b'', stderr=b'pg_rewind: error: could not open file '
+                           + b'"data/postgresql0/pg_xlog/000000010000000000000003": No such file')
     return 1
 
 
@@ -91,9 +91,10 @@ class TestRewind(BaseTestPostgresql):
                                              'Latest checkpoint location': '0/'})):
             self.r.rewind_or_reinitialize_needed_and_possible(self.leader)
 
-        with patch.object(Postgresql, 'is_running', Mock(return_value=True)):
-            with patch.object(MockCursor, 'fetchone', Mock(side_effect=[(0, 0, 1, 1, 0, 0, 0, 0, 0, None), Exception])):
-                self.r.rewind_or_reinitialize_needed_and_possible(self.leader)
+        with patch.object(Postgresql, 'is_running', Mock(return_value=True)),\
+                patch.object(MockCursor, 'fetchone',
+                             Mock(side_effect=[(0, 0, 1, 1, 0, 0, 0, 0, 0, None, None, None), Exception])):
+            self.r.rewind_or_reinitialize_needed_and_possible(self.leader)
 
     @patch.object(CancellableSubprocess, 'call', mock_cancellable_call)
     @patch.object(Postgresql, 'checkpoint', side_effect=['', '1'],)
@@ -222,7 +223,7 @@ class TestRewind(BaseTestPostgresql):
     @patch('patroni.postgresql.rewind.logger.info')
     def test_archive_ready_wals(self, mock_logger_info):
         with patch('os.listdir', Mock(side_effect=OSError)), \
-              patch.object(Postgresql, 'get_guc_value', Mock(side_effect=['on', 'command %f'])):
+             patch.object(Postgresql, 'get_guc_value', Mock(side_effect=['on', 'command %f'])):
             self.r._archive_ready_wals()
             mock_logger_info.assert_not_called()
 
@@ -232,19 +233,19 @@ class TestRewind(BaseTestPostgresql):
             'on', '',
         ]
         with patch.object(Postgresql, 'get_guc_value', Mock(side_effect=get_guc_value_res)):
-            for _ in range(len(get_guc_value_res)//2):
+            for _ in range(len(get_guc_value_res) // 2):
                 self.r._archive_ready_wals()
                 mock_logger_info.assert_not_called()
 
         with patch('os.listdir', Mock(return_value=['000000000000000000000000.ready'])):
             # successful archive_command call
-            with patch.object(CancellableSubprocess, 'call',  Mock(return_value=0)):
+            with patch.object(CancellableSubprocess, 'call', Mock(return_value=0)):
                 get_guc_value_res = [
                     'on', 'command %f',
                     'always', 'command %f',
                 ]
                 with patch.object(Postgresql, 'get_guc_value', Mock(side_effect=get_guc_value_res)):
-                    for _ in range(len(get_guc_value_res)//2):
+                    for _ in range(len(get_guc_value_res) // 2):
                         self.r._archive_ready_wals()
                         mock_logger_info.assert_called_once()
                         self.assertEqual(('Trying to archive %s: %s',
